"""Thread-aware browser tools that work with the browser session manager."""

import logging
from typing import Any, Dict, Optional, Type
from urllib.parse import urlparse

from langchain_core.runnables.config import RunnableConfig
from langchain_core.tools import BaseTool, ToolException
from pydantic import BaseModel, Field

from .browser_session_manager import BrowserSessionManager
from .utils import aget_current_page, get_current_page

logger = logging.getLogger(__name__)

class NavigateToolInput(BaseModel):
    """Input for NavigateTool."""
    url: str = Field(description="URL to navigate to")


class ClickToolInput(BaseModel):
    """Input for ClickTool."""
    selector: str = Field(
        description="CSS selector for the element to click on"
    )


class GetElementsToolInput(BaseModel):
    """Input for GetElementsTool."""
    selector: str = Field(
        description="CSS selector for elements to get"
    )


class ExtractTextToolInput(BaseModel):
    """Input for ExtractTextTool."""
    pass


class ExtractHyperlinksToolInput(BaseModel):
    """Input for ExtractHyperlinksTool."""
    pass


class NavigateBackToolInput(BaseModel):
    """Input for NavigateBackTool."""
    pass


class CurrentWebPageToolInput(BaseModel):
    """Input for CurrentWebPageTool."""
    pass


class ThreadAwareBaseTool(BaseTool):
    """Base class for thread-aware browser tools."""
    
    _session_manager: BrowserSessionManager
        
    def __init__(self, session_manager: BrowserSessionManager):
        """Initialize with a session manager."""
        super().__init__()
        self._session_manager = session_manager
        
    def get_thread_id(self, config: Optional[RunnableConfig] = None) -> str:
        """Extract thread ID from config."""
        thread_id = "default"

        if config and isinstance(config, dict):
            thread_id = config["configurable"]["thread_id"]
        
        return thread_id
        
    async def get_async_page(self, thread_id: str) -> Any:
        """Get or create a page for the specified thread."""
        browser = await self._session_manager.get_async_browser(thread_id)
        page = await aget_current_page(browser)
        return page
        
    def get_sync_page(self, thread_id: str) -> Any:
        """Get or create a page for the specified thread."""
        browser = self._session_manager.get_sync_browser(thread_id)
        page = get_current_page(browser)
        return page
        
    async def release_async_browser(self, thread_id: str) -> None:
        """Release the async browser session after use."""
        try:
            await self._session_manager.release_async_browser(thread_id)
            logger.debug(f"Released async browser for thread {thread_id}")
        except Exception as e:
            logger.warning(f"Error releasing async browser for thread {thread_id}: {e}")
    
    def release_sync_browser(self, thread_id: str) -> None:
        """Release the sync browser session after use."""
        try:
            self._session_manager.release_sync_browser(thread_id)
            logger.debug(f"Released sync browser for thread {thread_id}")
        except Exception as e:
            logger.warning(f"Error releasing sync browser for thread {thread_id}: {e}")


class ThreadAwareNavigateTool(ThreadAwareBaseTool):
    """Tool for navigating a browser to a URL with thread support."""

    name: str = "navigate_browser"
    description: str = "Navigate a browser to the specified URL"
    args_schema: Type[BaseModel] = NavigateToolInput
    
    def _run(
        self,
        url: str,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the sync tool."""
        thread_id = None
        try:
            logger.info(f"Passed config is {config}")

            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get page for this thread
            page = self.get_sync_page(thread_id)

            # Validate URL scheme
            parsed_url = urlparse(url)
            if parsed_url.scheme not in ("http", "https"):
                raise ValueError("URL scheme must be 'http' or 'https'")

            # Navigate to URL
            response = page.goto(url)
            status = response.status if response else "unknown"
            
            self.release_sync_browser(thread_id)
            
            return f"Navigating to {url} returned status code {status}"
        except Exception as e:
            if thread_id:
                try:
                    self.release_sync_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error navigating to {url}: {str(e)}")

    async def _arun(
        self,
        url: str,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the async tool."""
        thread_id = None
        try:
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get page for this thread
            page = await self.get_async_page(thread_id)

            # Validate URL scheme
            parsed_url = urlparse(url)
            if parsed_url.scheme not in ("http", "https"):
                raise ValueError("URL scheme must be 'http' or 'https'")

            # Navigate to URL
            response = await page.goto(url)
            status = response.status if response else "unknown"
            
            await self.release_async_browser(thread_id)
            
            return f"Navigating to {url} returned status code {status}"
        except Exception as e:
            if thread_id:
                try:
                    await self.release_async_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error navigating to {url}: {str(e)}")


class ThreadAwareClickTool(ThreadAwareBaseTool):
    """Tool for clicking on an element with the given CSS selector."""

    name: str = "click_element"
    description: str = "Click on an element with the given CSS selector"
    args_schema: Type[BaseModel] = ClickToolInput

    visible_only: bool = True
    """Whether to consider only visible elements."""
    playwright_strict: bool = False
    """Whether to employ Playwright's strict mode when clicking on elements."""
    playwright_timeout: float = 1_000
    """Timeout (in ms) for Playwright to wait for element to be ready."""

    def _selector_effective(self, selector: str) -> str:
        if not self.visible_only:
            return selector
        return f"{selector} >> visible=1"

    def _run(
        self,
        selector: str,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the sync tool."""
        thread_id = None
        try:
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get the current page
            page = self.get_sync_page(thread_id)

            # Click on the element
            selector_effective = self._selector_effective(selector=selector)
            from playwright.sync_api import TimeoutError as PlaywrightTimeoutError
            
            try:
                page.click(
                    selector_effective,
                    strict=self.playwright_strict,
                    timeout=self.playwright_timeout,
                )
            except PlaywrightTimeoutError:
                self.release_sync_browser(thread_id)
                return f"Unable to click on element '{selector}'"
            except Exception as click_error:
                self.release_sync_browser(thread_id)
                return f"Unable to click on element '{selector}': {str(click_error)}"
            
            self.release_sync_browser(thread_id)
            
            return f"Clicked element '{selector}'"
        except Exception as e:
            if thread_id:
                try:
                    self.release_sync_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error clicking on element: {str(e)}")

    async def _arun(
        self,
        selector: str,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the async tool."""
        thread_id = None
        try:
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get the current page
            page = await self.get_async_page(thread_id)

            # Click on the element
            selector_effective = self._selector_effective(selector=selector)
            from playwright.async_api import TimeoutError as PlaywrightTimeoutError
            
            try:
                await page.click(
                    selector_effective,
                    strict=self.playwright_strict,
                    timeout=self.playwright_timeout,
                )
            except PlaywrightTimeoutError:
                await self.release_async_browser(thread_id)
                return f"Unable to click on element '{selector}'"
            except Exception as click_error:
                await self.release_async_browser(thread_id)
                return f"Unable to click on element '{selector}': {str(click_error)}"
            
            await self.release_async_browser(thread_id)
            
            return f"Clicked element '{selector}'"
        except Exception as e:
            if thread_id:
                try:
                    await self.release_async_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error clicking on element: {str(e)}")


class ThreadAwareNavigateBackTool(ThreadAwareBaseTool):
    name: str = "navigate_back"
    description: str = "Navigate back to the previous page"
    args_schema: Type[BaseModel] = NavigateBackToolInput
    """Tool for navigating back in browser history."""

    def _run(
        self,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the sync tool."""
        thread_id = None
        try:
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get the current page
            page = self.get_sync_page(thread_id)

            # Navigate back
            try:
                page.go_back()
                self.release_sync_browser(thread_id)
                return "Navigated back to the previous page"
            except Exception as nav_error:
                self.release_sync_browser(thread_id)
                return f"Unable to navigate back: {str(nav_error)}"
        except Exception as e:
            if thread_id:
                try:
                    self.release_sync_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error navigating back: {str(e)}")

    async def _arun(
        self,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the async tool."""
        thread_id = None
        try:
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get the current page
            page = await self.get_async_page(thread_id)

            # Navigate back
            try:
                await page.go_back()
                await self.release_async_browser(thread_id)
                return "Navigated back to the previous page"
            except Exception as nav_error:
                await self.release_async_browser(thread_id)
                return f"Unable to navigate back: {str(nav_error)}"
        except Exception as e:
            if thread_id:
                try:
                    await self.release_async_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error navigating back: {str(e)}")


class ThreadAwareExtractTextTool(ThreadAwareBaseTool):
    name: str = "extract_text"
    description: str = "Extract all the text on the current webpage"
    args_schema: Type[BaseModel] = ExtractTextToolInput
    """Tool for extracting text from a webpage."""

    def _run(
        self,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the sync tool."""
        thread_id = None
        try:
            # Import BeautifulSoup
            try:
                from bs4 import BeautifulSoup
            except ImportError:
                raise ImportError(
                    "The 'beautifulsoup4' package is required to use this tool."
                    " Please install it with 'pip install beautifulsoup4'."
                )
                
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get the current page
            page = self.get_sync_page(thread_id)

            # Extract text
            content = page.content()
            soup = BeautifulSoup(content, "html.parser")
            
            self.release_sync_browser(thread_id)
            
            return soup.get_text(separator="\n").strip()
        except Exception as e:
            if thread_id:
                try:
                    self.release_sync_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error extracting text: {str(e)}")

    async def _arun(
        self,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the async tool."""
        thread_id = None
        try:
            # Import BeautifulSoup
            try:
                from bs4 import BeautifulSoup
            except ImportError:
                raise ImportError(
                    "The 'beautifulsoup4' package is required to use this tool."
                    " Please install it with 'pip install beautifulsoup4'."
                )
                
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get the current page
            page = await self.get_async_page(thread_id)

            # Extract text
            content = await page.content()
            soup = BeautifulSoup(content, "html.parser")
            
            await self.release_async_browser(thread_id)
            
            return soup.get_text(separator="\n").strip()
        except Exception as e:
            if thread_id:
                try:
                    await self.release_async_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error extracting text: {str(e)}")


class ThreadAwareExtractHyperlinksTool(ThreadAwareBaseTool):
    name: str = "extract_hyperlinks"
    description: str = "Extract all hyperlinks on the current webpage"
    args_schema: Type[BaseModel] = ExtractHyperlinksToolInput
    """Tool for extracting hyperlinks from a webpage."""

    def _run(
        self,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the sync tool."""
        thread_id = None
        try:
            # Import BeautifulSoup
            try:
                import json

                from bs4 import BeautifulSoup
            except ImportError:
                raise ImportError(
                    "The 'beautifulsoup4' package is required to use this tool."
                    " Please install it with 'pip install beautifulsoup4'."
                )
                
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get the current page
            page = self.get_sync_page(thread_id)

            # Extract hyperlinks
            content = page.content()
            soup = BeautifulSoup(content, "html.parser")
            links = []
            for link in soup.find_all("a", href=True):
                text = link.get_text().strip()
                href = link["href"]
                if href.startswith("http") or href.startswith("https"):
                    links.append({"text": text, "url": href})
            
            self.release_sync_browser(thread_id)
            
            if not links:
                return "No hyperlinks found on the current page."
            
            return json.dumps(links, indent=2)
        except Exception as e:
            if thread_id:
                try:
                    self.release_sync_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error extracting hyperlinks: {str(e)}")

    async def _arun(
        self,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the async tool."""
        thread_id = None
        try:
            # Import BeautifulSoup
            try:
                import json

                from bs4 import BeautifulSoup
            except ImportError:
                raise ImportError(
                    "The 'beautifulsoup4' package is required to use this tool."
                    " Please install it with 'pip install beautifulsoup4'."
                )
                
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get the current page
            page = await self.get_async_page(thread_id)

            # Extract hyperlinks
            content = await page.content()
            soup = BeautifulSoup(content, "html.parser")
            links = []
            for link in soup.find_all("a", href=True):
                text = link.get_text().strip()
                href = link["href"]
                if href.startswith("http") or href.startswith("https"):
                    links.append({"text": text, "url": href})
            
            await self.release_async_browser(thread_id)
            
            if not links:
                return "No hyperlinks found on the current page."
            
            return json.dumps(links, indent=2)
        except Exception as e:
            if thread_id:
                try:
                    await self.release_async_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error extracting hyperlinks: {str(e)}")


class ThreadAwareGetElementsTool(ThreadAwareBaseTool):
    name: str = "get_elements"
    description: str = "Get elements from the webpage using a CSS selector"
    args_schema: Type[BaseModel] = GetElementsToolInput
    """Tool for getting elements from a webpage."""

    def _run(
        self,
        selector: str,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the sync tool."""
        thread_id = None
        try:
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get the current page
            page = self.get_sync_page(thread_id)

            # Get elements
            elements = page.query_selector_all(selector)
            if not elements:
                self.release_sync_browser(thread_id)
                return f"No elements found with selector '{selector}'"
            
            elements_text = []
            for i, element in enumerate(elements):
                text = element.text_content()
                elements_text.append(f"Element {i+1}: {text.strip()}")
            
            self.release_sync_browser(thread_id)
            
            return "\n".join(elements_text)
        except Exception as e:
            if thread_id:
                try:
                    self.release_sync_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error getting elements: {str(e)}")

    async def _arun(
        self,
        selector: str,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the async tool."""
        thread_id = None
        try:
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get the current page
            page = await self.get_async_page(thread_id)

            # Get elements
            elements = await page.query_selector_all(selector)
            if not elements:
                await self.release_async_browser(thread_id)
                return f"No elements found with selector '{selector}'"
            
            elements_text = []
            for i, element in enumerate(elements):
                text = await element.text_content()
                elements_text.append(f"Element {i+1}: {text.strip()}")
            
            await self.release_async_browser(thread_id)
            
            return "\n".join(elements_text)
        except Exception as e:
            if thread_id:
                try:
                    await self.release_async_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error getting elements: {str(e)}")


class ThreadAwareCurrentWebPageTool(ThreadAwareBaseTool):
    name: str = "current_webpage"
    description: str = "Get information about the current webpage"
    args_schema: Type[BaseModel] = CurrentWebPageToolInput
    """Tool for getting information about the current webpage."""

    def _run(
        self,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the sync tool."""
        thread_id = None
        try:
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get the current page
            page = self.get_sync_page(thread_id)

            # Get information
            url = page.url
            title = page.title()
            
            self.release_sync_browser(thread_id)
            
            return f"URL: {url}\nTitle: {title}"
        except Exception as e:
            if thread_id:
                try:
                    self.release_sync_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error getting current webpage info: {str(e)}")

    async def _arun(
        self,
        config: RunnableConfig,
        **_,  # Ignore additional arguments
    ) -> str:
        """Use the async tool."""
        thread_id = None
        try:
            # Get thread ID from config
            thread_id = self.get_thread_id(config)

            # Get the current page
            page = await self.get_async_page(thread_id)

            # Get information
            url = page.url
            title = await page.title()
            
            await self.release_async_browser(thread_id)
            
            return f"URL: {url}\nTitle: {title}"
        except Exception as e:
            if thread_id:
                try:
                    await self.release_async_browser(thread_id)
                except Exception as release_error:
                    logger.warning(f"Error releasing browser: {release_error}")
            raise ToolException(f"Error getting current webpage info: {str(e)}")


def create_thread_aware_tools(
    session_manager: BrowserSessionManager,
) -> Dict[str, ThreadAwareBaseTool]:
    """
    Create thread-aware browser tools that use the session manager.
    
    Args:
        session_manager: The session manager to use for browser access
        
    Returns:
        Dictionary of thread-aware tools
    """
    # Import all required tool inputs
    try:
        import bs4  # noqa: F401
    except ImportError:
        import warnings
        warnings.warn(
            "The 'beautifulsoup4' package is required for extract_text and extract_hyperlinks tools."
            " Please install it with 'pip install beautifulsoup4'."
        )
    return {
        "navigate": ThreadAwareNavigateTool(session_manager=session_manager),
        "click": ThreadAwareClickTool(session_manager=session_manager),
        "navigate_back": ThreadAwareNavigateBackTool(session_manager=session_manager),
        "extract_text": ThreadAwareExtractTextTool(session_manager=session_manager),
        "extract_hyperlinks": ThreadAwareExtractHyperlinksTool(session_manager=session_manager),
        "get_elements": ThreadAwareGetElementsTool(session_manager=session_manager),
        "current_webpage": ThreadAwareCurrentWebPageTool(session_manager=session_manager),
    }



from __future__ import annotations

import hashlib
import json
import logging
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
    cast,
)

from langchain_core.caches import RETURN_VAL_TYPE, BaseCache
from langchain_core.embeddings import Embeddings
from langchain_core.load.dump import dumps
from langchain_core.load.load import loads
from langchain_core.outputs import Generation

from langchain_aws.vectorstores.inmemorydb import InMemoryVectorStore

logger = logging.getLogger(__file__)


def _hash(_input: str) -> str:
    """Use a deterministic hashing approach."""
    return hashlib.md5(_input.encode()).hexdigest()


def _dump_generations_to_json(generations: RETURN_VAL_TYPE) -> str:
    """Dump generations to json.

    Args:
        generations (RETURN_VAL_TYPE): A list of language model generations.

    Returns:
        str: Json representing a list of generations.

    Warning: would not work well with arbitrary subclasses of `Generation`
    """
    return json.dumps([generation.dict() for generation in generations])


def _load_generations_from_json(generations_json: str) -> RETURN_VAL_TYPE:
    """Load generations from json.

    Args:
        generations_json (str): A string of json representing a list of generations.

    Raises:
        ValueError: Could not decode json string to list of generations.

    Returns:
        RETURN_VAL_TYPE: A list of generations.

    Warning: would not work well with arbitrary subclasses of `Generation`
    """
    try:
        results = json.loads(generations_json)
        return [Generation(**generation_dict) for generation_dict in results]
    except json.JSONDecodeError:
        raise ValueError(
            f"Could not decode json to list of generations: {generations_json}"
        )


def _dumps_generations(generations: RETURN_VAL_TYPE) -> str:
    """
    Serialization for generic RETURN_VAL_TYPE, i.e. sequence of `Generation`

    Args:
        generations (RETURN_VAL_TYPE): A list of language model generations.

    Returns:
        str: a single string representing a list of generations.

    This function (+ its counterpart `_loads_generations`) rely on
    the dumps/loads pair with Reviver, so are able to deal
    with all subclasses of Generation.

    Each item in the list can be `dumps`ed to a string,
    then we make the whole list of strings into a json-dumped.
    """
    return json.dumps([dumps(_item) for _item in generations])


def _loads_generations(generations_str: str) -> Union[RETURN_VAL_TYPE, None]:
    """
    Deserialization of a string into a generic RETURN_VAL_TYPE
    (i.e. a sequence of `Generation`).

    See `_dumps_generations`, the inverse of this function.

    Args:
        generations_str (str): A string representing a list of generations.

    Compatible with the legacy cache-blob format
    Does not raise exceptions for malformed entries, just logs a warning
    and returns none: the caller should be prepared for such a cache miss.

    Returns:
        RETURN_VAL_TYPE: A list of generations.
    """
    try:
        generations = [loads(_item_str) for _item_str in json.loads(generations_str)]
        return generations
    except (json.JSONDecodeError, TypeError):
        # deferring the (soft) handling to after the legacy-format attempt
        pass

    try:
        gen_dicts = json.loads(generations_str)
        # not relying on `_load_generations_from_json` (which could disappear):
        generations = [Generation(**generation_dict) for generation_dict in gen_dicts]
        logger.warning(
            f"Legacy 'Generation' cached blob encountered: '{generations_str}'"
        )
        return generations
    except (json.JSONDecodeError, TypeError):
        logger.warning(
            f"Malformed/unparsable cached blob encountered: '{generations_str}'"
        )
        return None


class InMemorySemanticCache(BaseCache):
    """Cache that uses MemoryDB as a vector-store backend."""

    # TODO - implement a TTL policy in MemoryDB

    DEFAULT_SCHEMA = {
        "content_key": "prompt",
        "text": [
            {"name": "prompt"},
            {"name": "return_val"},
            {"name": "llm_string"},
        ],
    }

    def __init__(
        self, redis_url: str, embedding: Embeddings, score_threshold: float = 0.2
    ):
        """Initialize by passing in the `init` GPTCache func

        Args:
            redis_url (str): URL to connect to MemoryDB.
            embedding (Embedding): Embedding provider for semantic encoding and search.
            score_threshold (float, 0.2):

        Example:

        .. code-block:: python

            from langchain_core.globals import set_llm_cache

            from langchain_aws.cache import InMemorySemanticCache

            set_llm_cache(InMemorySemanticCache(
                redis_url="redis://localhost:6379",
                embedding=OpenAIEmbeddings()
            ))

        """
        self._cache_dict: Dict[str, InMemoryVectorStore] = {}
        self.redis_url = redis_url
        self.embedding = embedding
        self.score_threshold = score_threshold

    def _index_name(self, llm_string: str) -> str:
        hashed_index = _hash(llm_string)
        return f"cache:{hashed_index}"

    def _get_llm_cache(self, llm_string: str) -> InMemoryVectorStore:
        index_name = self._index_name(llm_string)

        # return vectorstore client for the specific llm string
        if index_name in self._cache_dict:
            return self._cache_dict[index_name]

        # create new vectorstore client for the specific llm string
        try:
            self._cache_dict[index_name] = InMemoryVectorStore.from_existing_index(
                embedding=self.embedding,
                index_name=index_name,
                redis_url=self.redis_url,
                schema=cast(Dict, self.DEFAULT_SCHEMA),
            )
        except ValueError:
            inmemory = InMemoryVectorStore(
                embedding=self.embedding,
                index_name=index_name,
                redis_url=self.redis_url,
                index_schema=cast(Dict, self.DEFAULT_SCHEMA),
            )
            _embedding = self.embedding.embed_query(text="test")
            inmemory._create_index_if_not_exist(dim=len(_embedding))
            self._cache_dict[index_name] = inmemory

        return self._cache_dict[index_name]

    def clear(self, **kwargs: Any) -> None:
        """Clear semantic cache for a given llm_string."""
        index_name = self._index_name(kwargs["llm_string"])
        if index_name in self._cache_dict:
            self._cache_dict[index_name].drop_index(
                index_name=index_name, delete_documents=True, redis_url=self.redis_url
            )
            del self._cache_dict[index_name]

    def lookup(self, prompt: str, llm_string: str) -> Optional[RETURN_VAL_TYPE]:
        """Look up based on prompt and llm_string."""
        llm_cache = self._get_llm_cache(llm_string)
        generations: List = []
        # Read from a Hash
        results = llm_cache.similarity_search(
            query=prompt,
            distance_threshold=0.1,
        )
        if results:
            for document in results:
                try:
                    generations.extend(loads(document.metadata["return_val"]))
                except Exception:
                    logger.warning(
                        "Retrieving a cache value that could not be deserialized "
                        "properly. This is likely due to the cache being in an "
                        "older format. Please recreate your cache to avoid this "
                        "error."
                    )
                    # In a previous life we stored the raw text directly
                    # in the table, so assume it's in that format.
                    generations.extend(
                        _load_generations_from_json(document.metadata["return_val"])
                    )
        return generations if generations else None

    def update(self, prompt: str, llm_string: str, return_val: RETURN_VAL_TYPE) -> None:
        """Update cache based on prompt and llm_string."""
        for gen in return_val:
            if not isinstance(gen, Generation):
                raise ValueError(
                    "InMemorySemanticCache only supports caching of "
                    f"normal LLM generations, got {type(gen)}"
                )
        llm_cache = self._get_llm_cache(llm_string)

        metadata = {
            "llm_string": llm_string,
            "prompt": prompt,
            "return_val": dumps([g for g in return_val]),
        }
        llm_cache.add_texts(texts=[prompt], metadatas=[metadata])



import json
import logging
from typing import Dict, List, Optional, Tuple

from bedrock_agentcore.tools.code_interpreter_client import CodeInterpreter
from langchain_core.runnables.config import RunnableConfig
from langchain_core.tools import BaseTool, StructuredTool

logger = logging.getLogger(__name__)


class CodeInterpreterToolkit:
    """Toolkit for working with AWS code interpreter environment.

    This toolkit provides a set of tools for working with a remote code interpreter environment:

    * execute_code - Run code in various languages (primarily Python)
    * execute_command - Run shell commands
    * read_files - Read content of files in the environment
    * list_files - List files in directories
    * delete_files - Remove files from the environment
    * write_files - Create or update files
    * start_command_execution - Start long-running commands asynchronously
    * get_task - Check status of async tasks
    * stop_task - Stop running tasks

    The toolkit lazily initializes the code interpreter session on first use.
    It supports multiple threads by maintaining separate code interpreter sessions for each thread ID.

    Example:
        ```python

        import asyncio
        from langgraph.prebuilt import create_react_agent
        from langchain_aws.tools import create_code_interpreter_toolkit

        async def main():
            # Create and setup the code interpreter toolkit
            toolkit, code_tools = await create_code_interpreter_toolkit(region="us-west-2")

            # Create a ReAct agent using the code interpreter tools
            agent = create_react_agent(
                model="bedrock_converse:us.anthropic.claude-3-5-haiku-20241022-v1:0",
                tools=code_tools
            )

            # Create runnable config with thread ID
            config = {
                "configurable": {
                    "thread_id": "session123"
                }
            }

            # Invoke the agent with a specific task using thread ID
            result = await agent.ainvoke(
                "Create a simple Python function that calculates the factorial of a number.",
                config=config
            )

            # Cleanup when done
            await toolkit.cleanup()

            return result

        # Run the example
        asyncio.run(main())
        ```
    """  # noqa: E501

    def __init__(self, region: str = "us-west-2"):
        """
        Initialize the toolkit

        Args:
            region: AWS region for the code interpreter
        """
        self.region = region
        self._code_interpreters: Dict[str, CodeInterpreter] = {}
        self.tools: List[BaseTool] = []

    def get_tools(self) -> List[BaseTool]:
        """
        Get the list of code interpreter tools

        Returns:
            List of LangChain tools
        """
        return self.tools

    def get_tools_by_name(self) -> Dict[str, BaseTool]:
        """
        Get a dictionary of tools mapped by their names

        Returns:
            Dictionary of {tool_name: tool}
        """
        return {tool.name: tool for tool in self.tools}

    def _get_or_create_interpreter(
        self, config: RunnableConfig
    ) -> CodeInterpreter:
        """
        Get or create a code interpreter for a specific config

        The config is expected to have a 'configurable' with 
        'thread_id', otherwise it creates a session with 'default'
        thread ID.

        Args:
            config: Runnable config that may contain a thread_id

        Returns:
            CodeInterpreter instance for the specified thread
        """
        # Extract thread ID from config if available
        thread_id = _get_thread_id(config)

        if thread_id in self._code_interpreters:
            return self._code_interpreters[thread_id]

        # Create a new code interpreter for this thread
        code_interpreter = CodeInterpreter(region=self.region)
        code_interpreter.start()
        logger.info(
            f"Started code interpreter with session_id:{code_interpreter.session_id} for thread:{thread_id}"  # noqa: E501
        )

        # Store the interpreter
        self._code_interpreters[thread_id] = code_interpreter
        return code_interpreter

    async def _setup(self) -> List[BaseTool]:
        """
        Setup and initialize code execution tools.

        Returns:
            List of LangChain tools for code execution
        """
        if self.tools:
            return self.tools

        # Create the basic tools for code execution - this doesn't initialize any code interpreter yet
        self.tools = self._create_tools()

        # Return the list of tools
        return self.tools

    def _create_tools(self) -> List[BaseTool]:
        """
        Create LangChain tools for code execution

        Returns:
            List of code execution tools
        """
        tools = []

        # Execute code tool
        execute_code_tool = StructuredTool.from_function(
            name="execute_code",
            func=self._execute_code,
        )
        tools.append(execute_code_tool)

        # Execute command tool
        execute_command_tool = StructuredTool.from_function(
            name="execute_command",
            func=self._execute_command,
        )
        tools.append(execute_command_tool)

        # Read files tool
        read_files_tool = StructuredTool.from_function(
            name="read_files",
            func=self._read_files,
        )
        tools.append(read_files_tool)

        # List files tool
        list_files_tool = StructuredTool.from_function(
            name="list_files",
            func=self._list_files,
        )
        tools.append(list_files_tool)

        # Remove files tool
        delete_files_tool = StructuredTool.from_function(
            name="delete_files",
            func=self._remove_files,
        )
        tools.append(delete_files_tool)

        # Write files tool
        write_files_tool = StructuredTool.from_function(
            name="write_files",
            func=self._write_files,
        )
        tools.append(write_files_tool)

        # Start command execution tool
        start_command_tool = StructuredTool.from_function(
            name="start_command_execution",
            func=self._start_command_execution,
        )
        tools.append(start_command_tool)

        # Get task status tool
        get_task_tool = StructuredTool.from_function(
            name="get_task",
            func=self._get_task,
        )
        tools.append(get_task_tool)

        # Stop task tool
        stop_task_tool = StructuredTool.from_function(
            name="stop_task",
            func=self._stop_task,
        )
        tools.append(stop_task_tool)

        return tools

    def _execute_code(
        self,
        code: str,
        config: RunnableConfig,
        language: str = "python",
        clearContext: bool = False,
    ) -> str:
        """
        Executes code in the AWS code interpreter environment

        Args:
            code:
                Code to execute
            language:
                Programming language, default is python
            clearContext:
                Whether to clear execution context, default is False
            config:
                Runnable config that may contain a thread_id

        Returns:
            String containing execution results
        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="executeCode",
            params={"code": code, "language": language, "clearContext": clearContext},
        )

        return _extract_output_from_stream(response)

    def _execute_command(
        self, command: str, config: RunnableConfig
    ) -> str:
        """
        Execute a command synchronously

        Args:
            command: Command to execute
            config: Runnable config that may contain a thread_id

        Returns:
            String containing execution results
        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="executeCommand", params={"command": command}
        )

        return _extract_output_from_stream(response)

    def _read_files(
        self, paths: List[str], config: RunnableConfig
    ) -> str:
        """
        Read content of files

        Args:
            paths: List of file paths to read
            config: Runnable config that may contain a thread_id

        Returns:
            String containing file contents
        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(method="readFiles", params={"paths": paths})

        return _extract_output_from_stream(response)

    def _list_files(
        self, config: RunnableConfig, directory_path: str = ""
    ) -> str:
        """
        List files in a directory

        Args:
            directory_path: Path to the directory to list, defaults to current directory
            config: Runnable config that may contain a thread_id

        Returns:
            String containing list of files
        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="listFiles", params={"directoryPath": directory_path}
        )

        return _extract_output_from_stream(response)

    def _remove_files(
        self, paths: List[str], config: RunnableConfig
    ) -> str:
        """
        Remove files from the system

        Args:
            paths: List of file paths to remove
            config: Runnable config that may contain a thread_id

        Returns:
            String containing removal result
        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="removeFiles", params={"paths": paths}
        )

        return _extract_output_from_stream(response)

    def _write_files(
        self, files: List[Dict[str, str]], config: RunnableConfig
    ) -> str:
        """
        Writes file content to the specified path in code env

        Is limited to writing files in the current working dir
        of the code interpreter environment. Absolute paths such
        as beginning with / are not allowed, only paths relative
        to current working dir, e.g., file.txt or dir/file.txt are
        valid paths.

        Args:
            files: List of dictionaries with path and text fields
            config: Runnable config that may contain a thread_id

        Returns:
            String containing write results
        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="writeFiles", params={"content": files}
        )

        return _extract_output_from_stream(response)

    def _start_command_execution(
        self, command: str, config: RunnableConfig
    ) -> str:
        """
        Start a long-running command asynchronously

        Args:
            command: Command to execute
            config: Runnable config that may contain a thread_id

        Returns:
            String containing task ID and status
        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="startCommandExecution", params={"command": command}
        )

        return _extract_output_from_stream(response)

    def _get_task(self, task_id: str, config: RunnableConfig) -> str:
        """
        Get status of an async task

        Args:
            task_id: ID of the task to check
            config: Runnable config that may contain a thread_id

        Returns:
            String containing task status
        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(method="getTask", params={"taskId": task_id})

        return _extract_output_from_stream(response)

    def _stop_task(self, task_id: str, config: RunnableConfig) -> str:
        """
        Stop a running task

        Args:
            task_id: ID of the task to stop
            config: Runnable config that may contain a thread_id

        Returns:
            String containing stop result
        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="stopTask", params={"taskId": task_id}
        )

        return _extract_output_from_stream(response)

    async def cleanup(self, thread_id: str = None) -> None:
        """Clean up resources

        Args:
            thread_id: Optional thread ID to clean up. If None, cleans up all sessions.
        """
        if thread_id:
            # Clean up a specific thread's session
            if thread_id in self._code_interpreters:
                try:
                    self._code_interpreters[thread_id].stop()
                    del self._code_interpreters[thread_id]
                    logger.info(
                        f"Code interpreter session for thread {thread_id} cleaned up"
                    )
                except Exception as e:
                    logger.warning(
                        f"Error stopping code interpreter for thread {thread_id}: {e}"
                    )
        else:
            # Clean up all sessions
            thread_ids = list(self._code_interpreters.keys())
            for tid in thread_ids:
                try:
                    self._code_interpreters[tid].stop()
                except Exception as e:
                    logger.warning(
                        f"Error stopping code interpreter for thread {tid}: {e}"
                    )

            self._code_interpreters = {}
            logger.info("All code interpreter sessions cleaned up")


async def create_code_interpreter_toolkit(
    region: str = "us-west-2",
) -> Tuple[CodeInterpreterToolkit, List[BaseTool]]:
    """
    Create and setup a CodeInterpreterToolkit

    Args:
        region: AWS region for code interpreter

    Returns:
        Tuple of (toolkit, tools)
    """
    toolkit = CodeInterpreterToolkit(region=region)
    # Create tools without immediately initializing the code interpreter
    tools = await toolkit._setup()
    # Code interpreter will be initialized lazily when first used
    return toolkit, tools


def _get_thread_id(config: Optional[RunnableConfig] = None):
    thread_id = "default"

    if config and isinstance(config, dict):
        thread_id = config["configurable"]["thread_id"]
    
    return thread_id

def _extract_output_from_stream(response):
    """
    Extract output from code interpreter response stream

    Args:
        response: Response from code interpreter execution

    Returns:
        Extracted output as string
    """
    output = []
    for event in response["stream"]:
        if "result" in event:
            result = event["result"]
            for content_item in result["content"]:
                if content_item["type"] == "text":
                    output.append(content_item["text"])
                if content_item["type"] == "resource":
                    resource = content_item["resource"]
                    if "text" in resource:
                        file_path = resource["uri"].replace("file://", "")
                        file_content = resource["text"]
                        output.append(f"==== File: {file_path} ====\n{file_content}\n")
                    else:
                        output.append(json.dumps(resource))

    return "\n".join(output)



from .browser_toolkit import create_browser_toolkit
from .code_interpreter_toolkit import create_code_interpreter_toolkit

__all__ = [
    "create_browser_toolkit",
    "create_code_interpreter_toolkit"
]


"""Toolkit for navigating web with AWS browser with thread support."""

import logging
from typing import Dict, List, Tuple

from langchain_core.tools import BaseTool

from .browser_session_manager import BrowserSessionManager
from .browser_tools import create_thread_aware_tools

logger = logging.getLogger(__name__)


class BrowserToolkit:
    """Toolkit for navigating web with AWS browser with thread support.

    This toolkit provides a set of tools for working with a remote browser
    and supports multiple threads by maintaining separate browser sessions
    for each thread ID. Browsers are created lazily only when needed.

    Example:
        ```python

        import asyncio
        from langgraph.prebuilt import create_react_agent
        from langchain_aws.tools import create_browser_toolkit

        async def main():
            # Create and setup the browser toolkit
            toolkit, browser_tools = create_browser_toolkit(region="us-west-2")

            # Create a ReAct agent using the browser tools
            agent = create_react_agent(
                model="bedrock_converse:us.anthropic.claude-3-5-haiku-20241022-v1:0",
                tools=browser_tools
            )

            # Create runnable config with thread ID
            config = {
                "configurable": {
                    "thread_id": "session123"
                }
            }

            # Invoke the agent with a specific task using thread ID
            result = await agent.ainvoke(
                "Navigate to https://www.example.com and tell me the main heading on the page.",
                config=config
            )

            # Clean up browser resources when done
            await toolkit.cleanup()

            return result

        # Run the example
        asyncio.run(main())
        ```
    """

    def __init__(self, region: str = "us-west-2"):
        """
        Initialize the toolkit

        Args:
            region: AWS region for the browser client
        """
        self.region = region
        self.session_manager = BrowserSessionManager(region=region)
        self.tools: List[BaseTool] = []
        self._setup_tools()

    def _setup_tools(self) -> None:
        """Initialize tools without creating any browsers."""
        # Create thread-aware tools with the session manager
        # Browsers will be created lazily when tools are used
        tools_dict = create_thread_aware_tools(self.session_manager)
        self.tools = list(tools_dict.values())

    def get_tools(self) -> List[BaseTool]:
        """
        Get the list of thread-aware browser tools

        Returns:
            List of LangChain tools
        """
        return self.tools

    def get_tools_by_name(self) -> Dict[str, BaseTool]:
        """
        Get a dictionary of tools mapped by their names

        Returns:
            Dictionary of {tool_name: tool}
        """
        return {tool.name: tool for tool in self.tools}

    async def cleanup(self) -> None:
        """Clean up all browser sessions"""
        await self.session_manager.close_all_browsers()
        logger.info("All browser sessions cleaned up")


def create_browser_toolkit(
    region: str = "us-west-2",
) -> Tuple[BrowserToolkit, List[BaseTool]]:
    """
    Create a BrowserToolkit with thread support

    Args:
        region: AWS region for browser client

    Returns:
        Tuple of (toolkit, tools)
    """
    toolkit = BrowserToolkit(region=region)
    tools = toolkit.get_tools()
    return toolkit, tools


import os
import re
from abc import abstractmethod
from typing import Any, Dict, Generic, Iterator, List, Literal, Optional, TypeVar, Union

from botocore.exceptions import BotoCoreError, UnknownServiceError
from langchain_core.messages import AIMessage
from packaging import version
from pydantic import SecretStr

MESSAGE_ROLES = Literal["system", "user", "assistant"]
MESSAGE_FORMAT = Dict[Literal["role", "content"], Union[MESSAGE_ROLES, str]]

INPUT_TYPE = TypeVar(
    "INPUT_TYPE", bound=Union[str, List[str], MESSAGE_FORMAT, List[MESSAGE_FORMAT]]
)
OUTPUT_TYPE = TypeVar(
    "OUTPUT_TYPE",
    bound=Union[str, List[List[float]], MESSAGE_FORMAT, List[MESSAGE_FORMAT], Iterator],
)


class ContentHandlerBase(Generic[INPUT_TYPE, OUTPUT_TYPE]):
    """A handler class to transform input from LLM and BaseChatModel to a

    format that SageMaker endpoint expects.

    Similarly, the class handles transforming output from the
    SageMaker endpoint to a format that LLM & BaseChatModel class expects.
    """

    """
    Example:
        .. code-block:: python

            class ContentHandler(ContentHandlerBase):
                content_type = "application/json"
                accepts = "application/json"

                def transform_input(self, prompt: str, model_kwargs: Dict) -> bytes:
                    input_str = json.dumps({prompt: prompt, **model_kwargs})
                    return input_str.encode('utf-8')
                
                def transform_output(self, output: bytes) -> str:
                    response_json = json.loads(output.read().decode("utf-8"))
                    return response_json[0]["generated_text"]
    """

    content_type: Optional[str] = "text/plain"
    """The MIME type of the input data passed to endpoint"""

    accepts: Optional[str] = "text/plain"
    """The MIME type of the response data returned from endpoint"""

    @abstractmethod
    def transform_input(self, prompt: INPUT_TYPE, model_kwargs: Dict) -> bytes:
        """Transforms the input to a format that model can accept
        as the request Body. Should return bytes or seekable file
        like object in the format specified in the content_type
        request header.
        """

    @abstractmethod
    def transform_output(self, output: bytes) -> OUTPUT_TYPE:
        """Transforms the output from the model to string that
        the LLM class expects.
        """


def enforce_stop_tokens(text: str, stop: List[str]) -> str:
    """Cut off the text as soon as any stop words occur."""
    return re.split("|".join(stop), text, maxsplit=1)[0]


def anthropic_tokens_supported() -> bool:
    """Check if all requirements for Anthropic count_tokens() are met."""
    try:
        import anthropic
    except ImportError:
        return False

    if version.parse(anthropic.__version__) > version.parse("0.38.0"):
        return False

    try:
        import httpx

        if version.parse(httpx.__version__) > version.parse("0.27.2"):
            raise ImportError()
    except ImportError:
        raise ImportError("httpx<=0.27.2 is required.")

    return True


def _get_anthropic_client() -> Any:
    import anthropic

    return anthropic.Anthropic()


def get_num_tokens_anthropic(text: str) -> int:
    """Get the number of tokens in a string of text."""
    client = _get_anthropic_client()
    return client.count_tokens(text=text)


def get_token_ids_anthropic(text: str) -> List[int]:
    """Get the token ids for a string of text."""
    client = _get_anthropic_client()
    tokenizer = client.get_tokenizer()
    encoded_text = tokenizer.encode(text)
    return encoded_text.ids


def create_aws_client(
    service_name: str,
    region_name: Optional[str] = None,
    credentials_profile_name: Optional[str] = None,
    aws_access_key_id: Optional[SecretStr] = None,
    aws_secret_access_key: Optional[SecretStr] = None,
    aws_session_token: Optional[SecretStr] = None,
    endpoint_url: Optional[str] = None,
    config: Any = None,
):
    """Helper function to validate AWS credentials and create an AWS client.

    Args:
        service_name: The name of the AWS service to create a client for.
        region_name: AWS region name. If not provided, will try to get from environment variables.
        credentials_profile_name: The name of the AWS credentials profile to use.
        aws_access_key_id: AWS access key ID.
        aws_secret_access_key: AWS secret access key.
        aws_session_token: AWS session token.
        endpoint_url: The complete URL to use for the constructed client.
        config: Advanced client configuration options.
    Returns:
        boto3.client: An AWS service client instance.

    """

    try:
        import boto3

        region_name = (
            region_name
            or os.getenv("AWS_REGION")
            or os.getenv("AWS_DEFAULT_REGION")
        )

        client_params = {
            "service_name": service_name,
            "region_name": region_name,
            "endpoint_url": endpoint_url,
            "config": config,
        }
        client_params = {
            k: v for k, v in client_params.items() if v
        }

        needs_session = bool(
            credentials_profile_name or
            aws_access_key_id or
            aws_secret_access_key or
            aws_session_token
        )

        if not needs_session:
            return boto3.client(**client_params)

        if credentials_profile_name:
            session = boto3.Session(profile_name=credentials_profile_name)
        elif aws_access_key_id and aws_secret_access_key:
            session_params = {
                "aws_access_key_id": aws_access_key_id.get_secret_value(),
                "aws_secret_access_key": aws_secret_access_key.get_secret_value(),
            }
            if aws_session_token:
                session_params["aws_session_token"] = aws_session_token.get_secret_value()
            session = boto3.Session(**session_params)
        else:
            raise ValueError(
                "If providing credentials, both aws_access_key_id and "
                "aws_secret_access_key must be specified."
            )

        if not client_params.get("region_name") and session.region_name:
            client_params["region_name"] = session.region_name

        return session.client(**client_params)

    except UnknownServiceError as e:
        raise ModuleNotFoundError(
            f"Ensure that you have installed the latest boto3 package "
            f"that contains the API for `{service_name}`."
        ) from e
    except BotoCoreError as e:
        raise ValueError(
            "Could not load credentials to authenticate with AWS client. "
            "Please check that the specified profile name and/or its credentials are valid. "
            f"Service error: {e}"
        ) from e
    except Exception as e:
        raise ValueError(f"Error raised by service:\n\n{e}") from e


def thinking_in_params(params: dict) -> bool:
    """Check if the thinking parameter is enabled in the request."""
    return params.get("thinking", {}).get("type") == "enabled"


def trim_message_whitespace(messages: List[Any]) -> List[Any]:
    """Trim trailing whitespace from final AIMessage content."""
    if not messages or not isinstance(messages[-1], AIMessage):
        return messages

    last_message = messages[-1]

    if isinstance(last_message.content, str):
        trimmed = last_message.content.rstrip()
        if trimmed != last_message.content:
            last_message.content = trimmed
    elif isinstance(last_message.content, list):
        for j, block in enumerate(last_message.content):
            if isinstance(block, dict) and block.get("type") == "text" \
                and isinstance(block.get("text"), str):
                trimmed = block["text"].rstrip()
                if trimmed != block["text"]:
                    last_message.content[j]["text"] = trimmed
    
    return messages



"""Utilities for the Playwright browser tools."""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Tuple

if TYPE_CHECKING:
    from playwright.async_api import Browser as AsyncBrowser
    from playwright.async_api import Page as AsyncPage
    from playwright.sync_api import Browser as SyncBrowser
    from playwright.sync_api import Page as SyncPage


async def aget_current_page(browser: AsyncBrowser) -> AsyncPage:
    """
    Asynchronously get the current page of the browser.

    Args:
        browser: The browser (AsyncBrowser) to get the current page from.

    Returns:
        AsyncPage: The current page.
    """
    if not browser.contexts:
        context = await browser.new_context()
        return await context.new_page()
    context = browser.contexts[0]  # Assuming you're using the default browser context
    if not context.pages:
        return await context.new_page()
    # Assuming the last page in the list is the active one
    return context.pages[-1]


def get_current_page(browser: SyncBrowser) -> SyncPage:
    """
    Get the current page of the browser.
    Args:
        browser: The browser to get the current page from.

    Returns:
        SyncPage: The current page.
    """
    if not browser.contexts:
        context = browser.new_context()
        return context.new_page()
    context = browser.contexts[0]  # Assuming you're using the default browser context
    if not context.pages:
        return context.new_page()
    # Assuming the last page in the list is the active one
    return context.pages[-1]


from langchain_aws.graphs.neptune_graph import (
    BaseNeptuneGraph,
    NeptuneAnalyticsGraph,
    NeptuneGraph,
)
from langchain_aws.graphs.neptune_rdf_graph import NeptuneRdfGraph

__all__ = [
    "BaseNeptuneGraph",
    "NeptuneAnalyticsGraph",
    "NeptuneGraph",
    "NeptuneRdfGraph",
]
